/************************************************
** WARNING: will only work with disk that have an MBR
** NO LFN support
************************************************/
#include <avr32/io.h>
#include "macro.h"
#include "vfs.h"


.global fat16_open
.global fat16_read
.global fat16_close
.global fat16_ls


/************************************************************************
*** r3 = device structure
*** r7 = path
*** r9 = entry to start at
*** r6 = Max number of entries (12bytes per entry)
*** r8 = buffer to receive entries
*** RETURN: r12 = number of entries read
************************************************************************/
fat16_ls:
    stm     --sp,r0-r11,lr

    sub     sp,VFS_FILE_STRUCT_SIZE

    mov     r0,r8                                           // just to save  temporarily
    mov     r5,sp
    ld.w    r8,r3[VFS_DEV_TEMPBUF]
    rcall   fat16_getPartition
    cp      r12,0
    brne    fat16_ls_exit
    mov     r8,r0

    // check if we should search root
    ld.ub   r0,r7[3]
    mov     r5,sp
    cp      r0,0
    brne    fat16_ls_searchfolder

    /* search root */
    mov     r0,r6
    rcall   fat16_listroot
    sub     r12,r0,r12
    rjmp    fat16_ls_exit

fat16_ls_searchfolder:
    sub     r7,-3                                       // skip device name
    rcall   fat16_findfile
    cp      r12,2
    breq    2f
    cp      r12,1
    breq    1f
    SERIALPRINTF("\r\nFILE NOT FOUND\r\n")
    rjmp    fat16_ls_exit
1:  SERIALPRINTF("\r\nNOT A DIRECTORY\r\n")
    rjmp    fat16_ls_exit
2:  mov     r1,r6
    rcall   fat16_listfolder
    sub     r12,r1,r12

fat16_ls_exit:
    sub     sp,-(VFS_FILE_STRUCT_SIZE)
    ldm     sp++,r0-r11,pc


/************************************************************************
*** r0 = cluster
*** r3 = device structure
*** r9 = entry to start at
*** r5 = FILE struct
*** r6 = Max number of entries (12bytes per entry)
*** r8 = buffer to receive entries
*** RETURN: r12 = number of entries left to read
************************************************************************/
fat16_listfolder:
    stm     --sp,r0-r11,lr

    mov     r10,r8
    sub     r9,-1

lsfolderfgetblock:
    ld.w    r8,r3[VFS_DEV_TEMPBUF]
    rcall   fat16_readcluster
    
    // loop through all entries
    ld.w    r1,r5[VFS_FILE_FAT16_CLUSTERSIZE]
    lsl     r1,4                                        // blocks_per_sector * 16 entries
lsfolderfnextentry:
    ld.ub   r2,r8[0x0B]
    andl    r2,0x0F
    cp      r2,0x0F
    breq    lsfolderfskiptonextentry                             // Skip LFN entries
    ld.ub   r2,r8[0x00]
    cp      r2,0xE5
    breq    lsfolderfskiptonextentry
    cp      r2,0
    breq    lsfolder_exit
    cp      r9,0                                        // skip some entries
    breq    1f
    sub     r9,1
    brne    lsfolderfskiptonextentry
1:  ld.w    r2,r8[0]
    st.w    r10++,r2
    ld.w    r2,r8[4]
    st.w    r10++,r2
    mov     r2,'.'
    st.b    r10++,r2
    ld.w    r2,r8[8]
    st.w    r10++,r2
    mov     r2,0
    st.b    r10++,r2
    sub     r6,1
    breq    lsfolder_exit

lsfolderfskiptonextentry:
    sub     r8,-32
    sub     r1,1
    brne    lsfolderfnextentry

    rcall   fat16_getnextcluster
    mov     r0,r12
    mov     r12,0
    cp      r0,0xFFF0
    brlo    lsfolderfgetblock                                   // cluster is valid, search it

lsfolder_exit:
    mov     r12,r6
    ldm     sp++,r0-r11,pc



/************************************************************************
*** r3 = device structure
*** r9 = entry to start at
*** r5 = FILE struct
*** r6 = Max number of entries (14bytes per entry)
*** r8 = buffer to receive entries
*** RETURN: r12 = number of entries left to read
************************************************************************/
fat16_listroot:
    stm     --sp,r0-r11,lr
    ld.w    r0,r5[VFS_FILE_FAT16_ROOTDIRBLOCK]
    mov     r10,r8
    sub     r9,-1

lsrootgetblock:
    ld.w    r8,r3[VFS_DEV_TEMPBUF]
    mov     r7,r0
    mcall   r3[VFS_DEV_READ]                            // read sector into stack buffer

    /* loop through all entries */
    mov     r1,16                                       // 16 entries per blocks
lsrootnextentry:
    ld.ub   r2,r8[0x0B]
    andl    r2,0x0F
    cp      r2,0x0F
    breq    lsrootskiptonextentry                             // Skip LFN entries
    ld.ub   r2,r8[0x00]
    cp      r2,0xE5
    breq    lsrootskiptonextentry
    cp      r2,0
    breq    lsroot_exit
    cp      r9,0                                        // skip some entries
    breq    1f
    sub     r9,1
    brne    lsrootskiptonextentry
1:  ld.w    r2,r8[0]
    st.w    r10++,r2
    ld.w    r2,r8[4]
    st.w    r10++,r2
    mov     r2,'.'
    st.b    r10++,r2
    ld.w    r2,r8[8]
    st.w    r10++,r2
    mov     r2,0
    st.b    r10++,r2

    sub     r6,1
    breq    lsroot_exit

lsrootskiptonextentry:
    sub     r8,-32
    sub     r1,1
    brne    lsrootnextentry

    /*increase block number and exit if equal to datablock (r0=currentblock)*/
    ld.w    r8,r5[VFS_FILE_FAT16_DATABLOCK]
    sub     r0,-1
    cp      r0,r8
    brne    lsrootgetblock

lsroot_exit:
    mov     r12,r6
    ldm     sp++,r0-r11,pc













/************************************************************************
*** r5 = pointer to a FILE structure
************************************************************************/
fat16_close:
    stm     --sp,r0-r11,lr
    ldm     sp++,r0-r11,pc

/************************************************************************
*** r5 = pointer to a FILE structure
*** r11 = number of bytes to read
*** r8 = buffer which the data will be copied into
*** RETURN: r12 = number of bytes read
************************************************************************/
fat16_read:
    stm     --sp,r0-r11,lr

    /* make sure r11 does not specify a value that would read pass the end of file */
    ld.w    r0,r5[VFS_FILE_FAT16_CURRENT_OFFSET]

    ld.w    r10,r5[VFS_FILE_FAT16_FILE_SIZE]
    add     r9,r0,r11
    cp      r9,r10
    brls    1f
    sub     r11,r10,r0

    /* get initial data */
1:  ld.w    r1,r5[VFS_FILE_FAT16_CLUSTERSIZE]
    lsl     r4,r1,9                                            //clustersize*512 = clustersize in bytes
    divu    r0,r0,r4                                          // r0+1 (r1) becomes offset within cluster
    ld.w    r0,r5[VFS_FILE_FAT16_CURRENTCLUSTER]

    mov     r10,0
    mov     r9,r8
fat16_fread_next:
    cp      r0,0xFFEF
    brhi    fat16_fread_exit                                // end of file reached
    ld.w    r8,r3[VFS_DEV_TEMPBUF]                             
    rcall   fat16_readcluster
    add     r8,r1

    /* copy data to buffer */
    sub     r1,r4,r1                                        // bytes tl read in temp buffer -> r1
1:  ld.ub   r2,r8++                                         // read from temp
    st.b    r9++,r2                                         // write in user buffer
    sub     r10,-1                                          // total bytes read for this call
    cp      r10,r11                                         // number of requested bytes all read?
    breq    fat16_fread_exit                                //   yes. done reading, exit.
    sub     r1,1                                            // bytes left to read in buffer
    brne    1b                                              // if no more bytes to read, get next cluster

    /* get next cluster value */
    rcall   fat16_getnextcluster
    mov     r0,r12
    st.w    r5[VFS_FILE_FAT16_CURRENTCLUSTER],r0
    rjmp    fat16_fread_next

fat16_fread_exit:
    mov     r12,r10
    ld.w    r10,r5[VFS_FILE_FAT16_CURRENT_OFFSET]
    add     r10,r12
    st.w    r5[VFS_FILE_FAT16_CURRENT_OFFSET],r10
    ldm     sp++,r0-r11,pc


/*******************************************************
** r3 = device functions virtual table pointer
** r5 = FILE structure
** r8 = buffer to hold the partition sector
** RETURN r12=0 if success
********************************************************/
fat16_getPartition:
    stm     --sp,r0,r7,lr

    /* read first sector */
    mov     r7,0
    mcall   r3[VFS_DEV_READ]                            // read MBR

    /* validate that this is a fat16 partition */
    ld.ub   r7,r8[0x1BE+4]                              // partition type of partition 1
    cp      r7,06                                       // Fat16 partition?
    breq    1f
    SERIALPRINTF("Invalid partition type\r\n")
    mov     r12,1
    rjmp    fat16_getpartition_exit
1:
    /* find partition */
    ld.w    r7,r8[0x1BE+8]                              // LBA address of partition
    swap.b  r7                                          // convert from intel order
    mov     r0,r7                                      // save temporarily
    mcall   r3[VFS_DEV_READ]                            // read sector into stack buffer
    
    /* find fat sector */
    ld.uh   r7,r8[0x0E]                                 // number of reserved
    swap.bh r7                                          // convert from intel order
    add     r0,r7                                       // r0 now contains the block for FAT
    st.w    r5[VFS_FILE_FAT16_FATBLOCK],r0

    /* check sector size */
    ld.w    r7,r8[0x0A]
    lsr     r7,8
    andh    r7,0
    swap.bh r7
    mov     r12,2
    cp      r7,0x200
    brne    fat16_getpartition_exit                     // we only support 512bytes sectors right now    

    /* get root directory */
    ld.uh   r7,r8[0x16]
    swap.bh r7
    lsl     r7,1
    add     r7,r0
    st.w    r5[VFS_FILE_FAT16_ROOTDIRBLOCK],r7
    ld.ub   r0,r8[0x0D]
    mov     r12,3
    cp      r0,0x40
    brhi    fat16_getpartition_exit                     // we don't support more than 64 sectors per cluster right now
    st.w    r5[VFS_FILE_FAT16_CLUSTERSIZE],r0
 
    /* get data region */
    ld.w    r0,r8[0x10]                                 // wanna get half word att 0x11, but must use aligned address
    lsr     r0,8            
    andh    r0,0
    swap.bh r0
    lsr     r0,4
    add     r7,r7,r0                                     // rootdir+((numRootEntries*32)/512)
    st.w    r5[VFS_FILE_FAT16_DATABLOCK],r7

    mov     r12,0
fat16_getpartition_exit:
    ldm     sp++,r0,r7,pc

/*******************************************************
** r3 = device functions virtual table pointer
** r8 = buffer to hold data
** r5 = FILE structure
** r0= cluster number
** RETURN: r12 = 0 if sucess
********************************************************/
fat16_readcluster:
    stm     --sp,r0,r1,r8,r9,r7,lr

    /* calculate block number of requested cluster */
    sub     r0,2
    ld.w    r1,r5[VFS_FILE_FAT16_CLUSTERSIZE]
    mul     r0,r1
    ld.w    r7,r5[VFS_FILE_FAT16_DATABLOCK]
    add     r7,r0

    // read all sectors
1:  mcall   r3[VFS_DEV_READ]
    ld.w    r9,r3[VFS_DEV_BLKSIZE]
    add     r8,r9
    sub     r7,-1

    sub     r1,1
    brne    1b

    ldm     sp++,r0,r1,r8,r9,r7,pc

/*******************************************************
** Will take a filename and convert it in
**  the same format as on the filesystem (uppercase, padded, no dots for extension etc..)
** r7 = pointer filename
** r8 = buffer to hold new value
** RETURN: r12 = 0 if filename is valid, 1 if invalid
********************************************************/
fat16_convertfilename:
    stm     --sp,r1,r2,r7,r8,r9,lr
    mov     r12,0

    mov     r9,r8
    mov     r1,0x2020
    orh     r1,0x2020
    st.w    r8,r1
    st.w    r8[4],r1
    st.w    r8[8],r1                                // clear the buffer with spaces
    mov     r2,9                                    // no more than 8chars, but we decrease when entering loop                                 
1:  ld.ub   r1,r7++
    sub     r2,1
    breq    convertextention                        // if 8 chars were read, consider the rest as an extension
    cp.w    r1,'.'
    breq    convertextention                        // if we have a '.', extension starts 
    cp.w    r1,'/'                                  // end of filename, no extension
    breq    convertexit 
    cp.w    r1,0
    breq    convertexit                                     // end of filename, no extension
    cp      r1,'a'
    brlo    nouppercase
    cp      r1,'z'
    brhi    nouppercase
    sub     r1,32                                   // convert to uppercase
nouppercase:
    st.b    r8++,r1
    rjmp    1b
    
    /* handle the extension */
convertextention:
    cp      r1,'.'
    breq    1f                  // is this the extension or we just busted the 8char limit?
    mov     r12,1
    rjmp convertexit
1:  mov     r8,r9
    sub     r8,-8
    mov     r2,3
1:  ld.ub   r1,r7++
    cp.w    r1,0
    breq    convertexit 
    cp      r1,'a'
    brlo    extnouppercase
    cp      r1,'z'
    brhi    extnouppercase
    sub     r1,32
extnouppercase:
    st.b    r8++,r1
    sub     r2,1
    brne    1b
convertexit:
    ldm     sp++,r1,r2,r7,r8,r9,pc



/*******************************************************
** r3 = device functions virtual table pointer
** r7 = pointer filename
** r5 = pointer to empty FILE structure
** RETURN: r12 = 0=not-found, 1=isfile, 2=isdir
** RETURN: r0 = cluster
********************************************************/
fat16_findfile:
    stm     --sp,r1,r7,r8,r9,lr
    sub     sp,12                                           // buffer to hold converted filename

    /*convert filename to be like it would be found on FS. r7 already hold the filename pointer */
    mov     r8,sp
    rcall   fat16_convertfilename
    cp      r12,0
    brne    notfound

    /* find the root entry */
    mov     r9,r7
    mov     r7,sp
    rcall   fat16_findinroot
    mov     r7,r9
    cp      r12,0
    breq    fat16_findfile_exit 
    
1:  ld.ub   r9,r7++
    cp      r9,'/'
    breq    isdir
    cp      r9,0
    breq    isfile
    rjmp    1b
isdir:
    cp      r12,2
    brne    notfound

    ld.ub   r8,r7
    cp      r8,0                                            // end of path
    breq    fat16_findfile_exit    

    /* Search in subdirectory */
    mov     r8,sp
    rcall   fat16_convertfilename
    cp      r12,0
    brne    notfound
    mov     r9,r7
    mov     r7,sp
    rcall   fat16_findinfolder
    mov     r7,r9
    cp      r12,0
    breq    fat16_findfile_exit
1:  ld.ub   r9,r7++
    cp      r9,'/'
    breq    isdir                                           // search in next directory
    cp      r9,0
    breq    isfile                                          //  file is found
    rjmp    1b
isfile:
    cp      r12,1
    brne    notfound
    st.w    r5[VFS_FILE_FAT16_FILE_SIZE],r1                 // save file size
    st.w    r5[VFS_FILE_FAT16_CURRENTCLUSTER],r0            // save first cluster
    rjmp    fat16_findfile_exit
notfound:
    mov     r12,0
fat16_findfile_exit:
    sub     sp,-12
    ldm     sp++,r1,r7,r8,r9,pc


/*******************************************************
** r5 = pointer FILE structure
** r3 = pointer to device info structure
** r7 = pointer filename
** RETURN: r12 = 0 = not found, 1=is a file, 2=is a directory
** RETURN: r0 = cluster
** RETURN: r1 = filesize
********************************************************/
fat16_findinroot:
    stm     --sp,r2-r4,r7-r9,lr

    /* find the file in root directory */
    mov     r12,0
    ld.w    r0,r5[VFS_FILE_FAT16_ROOTDIRBLOCK]
getblock:
    mov     r1,r7                   // store temporarily
    ld.w    r8,r3[VFS_DEV_TEMPBUF]
    mov     r7,r0
    mcall   r3[VFS_DEV_READ]                            // read sector into stack buffer
    mov     r7,r1

    // loop through all entries
    mov     r1,16                                       // 16 entries per blocks
nextentry:
    ld.ub   r2,r8[0x0B]
    andl    r2,0x0F
    cp      r2,0x0F
    breq    skiptonextentry                             // Skip LFN entries
    ld.ub   r2,r8[0x00]
    cp      r2,0xE5
    breq    skiptonextentry
    cp      r2,0
    breq    skiptonextentry

    /* compare filename */
    ld.w    r2,r8[0]
    ld.w    r4,r7[0]
    cp      r2,r4
    brne    skiptonextentry
    ld.w    r2,r8[4]
    ld.w    r4,r7[4]
    cp      r2,r4
    brne    skiptonextentry
    ld.w    r2,r8[8]
    ld.w    r4,r7[8]
    lsr     r2,8                                        // last byte is garbage
    lsr     r4,8                                        // last byte is garbage
    cp      r2,r4
    brne    skiptonextentry
    mov     r12,1
    ld.ub   r0,r8[0x0B]
    bld     r0,4
    acr     r12
    ld.uh   r0,r8[0x1A]                                 // starting cluster
    swap.bh r0
    ld.w    r1,r8[0x1C]
    swap.b  r1
    rjmp    fat16_findinroot_exit
skiptonextentry:
    sub     r8,-32
    sub     r1,1
    brne    nextentry

    /*increase block number and exit if equal to datablock (r0=currentblock)*/
    ld.w    r8,r5[VFS_FILE_FAT16_DATABLOCK]
    sub     r0,-1
    cp      r0,r8
    brne    getblock
    mov     r12,0

fat16_findinroot_exit:
    ldm     sp++,r2-r4,r7-r9,pc


/*******************************************************
** r5 = pointer FILE structure
** r3 = pointer to device info structure
** r7 = pointer filename
** r0 = cluster
** RETURN: r12 = 0 = not found, 1=is a file, 2=is a directory
** RETURN: r0 = cluster
********************************************************/
fat16_findinfolder:
    stm     --sp,r2-r4,r7-r9,lr

fgetblock:
    ld.w    r8,r3[VFS_DEV_TEMPBUF]
    rcall   fat16_readcluster

    // loop through all entries
    ld.w    r1,r5[VFS_FILE_FAT16_CLUSTERSIZE]
    lsl     r1,4                                        // blocks_per_sector * 16 entries
fnextentry:
    ld.ub   r2,r8[0x0B]
    andl    r2,0x0F
    cp      r2,0x0F
    breq    fskiptonextentry                             // Skip LFN entries
    ld.ub   r2,r8[0x00]
    cp      r2,0xE5
    breq    fskiptonextentry
    cp      r2,0
    breq    fskiptonextentry

    /* compare filename */
    ld.w    r2,r8[0]
    ld.w    r4,r7[0]
    cp      r2,r4
    brne    fskiptonextentry
    ld.w    r2,r8[4]
    ld.w    r4,r7[4]
    cp      r2,r4
    brne    fskiptonextentry
    ld.w    r2,r8[8]
    ld.w    r4,r7[8]
    lsr     r2,8                                        // last byte is garbage
    lsr     r4,8                                        // last byte is garbage
    cp      r2,r4
    brne    fskiptonextentry
    mov     r12,1
    ld.ub   r0,r8[0x0B]
    bld     r0,4
    acr     r12
    ld.uh   r0,r8[0x1A]                                 // starting cluster
    swap.bh r0
    ld.w    r1,r8[0x1C]
    swap.b  r1
    rjmp    fat16_findinfolder_exit
fskiptonextentry:
    sub     r8,-32
    sub     r1,1
    brne    fnextentry

    rcall   fat16_getnextcluster
    mov     r0,r12
    mov     r12,0
    cp      r0,0xFFF0
    brlo    fgetblock                                   // cluster is valid, search it

fat16_findinfolder_exit:
    ldm     sp++,r2-r4,r7-r9,pc


/*******************************************************
** r3 = device functions virtual table pointer
** r0 = cluster
** r5 = pointer to empty FILE structure
** RETURN: r12 = next cluster
********************************************************/
fat16_getnextcluster:
    stm     --sp,r6,r7,r8,r9,lr

    lsl     r8,r0,1                                        //cluster are 16bit wide. multiply by 2 to get absolute position.
    lsr     r7,r8,9                                        //divide by 512 to find block
    sub     r6,r8,r7<<9                                    //r6=offset in block
    ld.w    r9,r5[VFS_FILE_FAT16_FATBLOCK]
    add     r7,r9
    ld.w    r8,r3[VFS_DEV_TEMPBUF]
    mcall   r3[VFS_DEV_READ]
    add     r8,r6

    ld.uh   r12,r8
    swap.bh r12

    ldm     sp++,r6,r7,r8,r9,pc



/*******************************************************
** r3 = device functions virtual table pointer
** r7 = pointer filename
** r5 = pointer to empty FILE structure
** RETURN: r12 = 0 if sucess
********************************************************/
fat16_open:
    stm     --sp,r0,r1,r4,r8,r9,r11,lr

    /* read first sector. r5 already contains FILE pointer */
    ld.w    r8,r3[VFS_DEV_TEMPBUF]
    rcall   fat16_getPartition
    cp      r12,0
    brne    fat16_getfirstblock_exit

    sub     r7,-3                                       // skip device name
    rcall   fat16_findfile
    cp      r12,0
    brne    1f
    SERIALPRINTF("\r\nFILE NOT FOUND\r\n")
    rjmp    fat16_getfirstblock_exit
1:
    mov     r9,0
    st.w    r5[VFS_FILE_FAT16_CURRENT_OFFSET],r9
    mov     r12,0

fat16_getfirstblock_exit:
    ldm     sp++,r0,r1,r4,r8,r9,r11,pc


