#include <avr32/io.h>
#include "macro.h"
#include "vfs.h"
#define RESETSTATREG    ((1 << AVR32_SR_M_OFFSET)|(1 << AVR32_SR_EM_OFFSET)|(1<<AVR32_SR_GM_OFFSET))
#define RESETCPUCR  ((1 << AVR32_CPUCR_BI_OFFSET)|(1 << AVR32_CPUCR_BE_OFFSET)|\
                    (1<<AVR32_CPUCR_RE_OFFSET)|(1<<AVR32_CPUCR_IBE_OFFSET)|(1<<AVR32_CPUCR_IEE_OFFSET))

.global main

.extern sdram_init
.extern initInterrupts
.extern initSerial
.extern recvSerial


.section	.text
.align	1

main:
    /* stack starts at 0x24001FFF downto 0x24000000 which is the first 2K of SRAM */
    /* after paging is enabled we will access SRAM from the P3 segment by remapping the address */
	lddpc 	sp,.tempstackBase

    ////////////////////////////////////
    // reset the system
    ////////////////////////////////////
    /* Reset the Status Register */
    mov     r0,lo(RESETSTATREG)
    orh     r0,hi(RESETSTATREG)
    mtsr    AVR32_SR,r0

    /* Reset CPUCR and invalidate the BTB */
    mov     r2,RESETCPUCR
    mtsr    AVR32_CPUCR,r2

    /* Reset the MMU to default settings */
    mov     r0, ((1<< AVR32_MMUCR_S) | (1<<AVR32_MMUCR_I))
    mtsr    AVR32_MMUCR,r0


    ////////////////////////////////////
    // initialize
    ////////////////////////////////////
    /* init power manager */
    rcall   init_pm

    /* initialize EVBA */
    rcall   initEVBA
    
    /* init serial port */
    mov     r8,1
    mov     r7,7            // 115200
    rcall   serialInit

    /* initialize SDRAM */
    rcall   sdram_init
    WAIT(0x800000)

    /* initialize dataflash */
    rcall   dataflash_init

    /* initialize PIOA */
    mov     r6,0x80000
    lddpc   r8,.pioa
    st.w    r8[AVR32_PIO_PER],r6
    st.w    r8[AVR32_PIO_OER],r6
    st.w    r8[AVR32_PIO_IDR],r6
    st.w    r8[AVR32_PIO_PUDR],r6

    /* init interrupts */
    rcall   initInterrupts

    /* Init MMU */
    rcall   init_mmu

    /* init MCI */
    rcall   mci_init

    /* init MACB 0*/
    mov     r5,0
    rcall   init_macb

    ////////////////////////////////////
    // main loop
    ////////////////////////////////////

    //TODO: should determine real size
    
    // The application has been loaded (using u-boot) at 0x80110000
    mov     r8,0x0000
    orh     r8,0x8011
    mov     r9,0x0FFF       //TODO: must determine real size using elf header
    orh     r9,0x8011
    rcall   LoadProcessFromMem

    /* start scheduler */
    rcall   init_scheduler

/*SERIALPRINTF("listening\r\n")
1:mov r5,0
lddpc r9,.dataflashBuffer
rcall macb_getframe
cp r12,0
breq 1b
lddpc r8,.dataflashBuffer
mov r7,r12
rcall debug_showMem
rjmp 1b*/

ml1:
    SERIALPRINTF("mainloop\r\n")
    WAIT(0x8000000)
    rjmp ml1





/**********************************************************
*** Constants (WARNING: must be word aligned)
**********************************************************/
safe: rjmp safe
.testFileName: .ascii "/1/tsapit~2.txt\0"
.align 4
.testFileName2: .ascii "/1/dir2/crap.txt\0"
.align 4
.testFileName3: .ascii "/1/dir2/\0"
.align 4
.testFileName4: .ascii "/1/dir1\0"
.align 4
.dataflashBuffer: .int 0xB0010000
.kernelHeap: .int 0x04000000
.tempstackBase: .int 0x24000FFC // downto 0x24000000, thus 4K of stack
.pioa:      .int AVR32_PIOA_ADDRESS
.sdramc:    .int AVR32_SDRAMC_ADDRESS
.hmatrix:   .int AVR32_HMATRIX_ADDRESS

