#include <avr32/io.h>
#include "macro.h"

.global dtlbMissWrite
.global instructionAddress
.global itlbProtection
.global dataAddressRead
.global dataAddressWrite
.global dtlbProtectionRead
.global dtlbProtectionWrite
.global dtlbModified
.global itlbMiss
.global dtlbMissRead
.global init_mmu
.global PageAlloc
.global PageDealloc
.global PageDeallocAll
.global malloc
.global PageAllocCode
.global PageAllocData


instructionAddress:
    SERIALPRINTF("MMU ERROR1 from: ")
    mfsr    r8,AVR32_RAR_EX
    rcall   debug_showRegister
    1: rjmp 1b
itlbProtection:
    SERIALPRINTF("MMU ERROR2 from: ")
    mfsr    r8,AVR32_RAR_EX
    rcall   debug_showRegister
    1: rjmp 1b
dataAddressRead:
    SERIALPRINTF("MMU ERROR3 from: ")
    mfsr    r8,AVR32_RAR_EX
    rcall   debug_showRegister
    1: rjmp 1b
dataAddressWrite:
    SERIALPRINTF("MMU: ADRESS WRITE AT: ")
    mfsr    r8,AVR32_TLBEAR
    rcall   debug_showRegister
    SERIALPRINTF(" FROM: ")
    mfsr    r8,AVR32_RAR_EX
    rcall   debug_showRegister

    1: rjmp 1b
dtlbProtectionRead:
    SERIALPRINTF("MMU ERROR5 from: ")
    mfsr    r8,AVR32_RAR_EX
    rcall   debug_showRegister
    1: rjmp 1b
dtlbProtectionWrite:
    SERIALPRINTF("MMU ERROR6 from: ")
    mfsr    r8,AVR32_RAR_EX
    rcall   debug_showRegister
    1: rjmp 1b
dtlbModified:
    SERIALPRINTF("MMU ERROR7 from: ")
    mfsr    r8,AVR32_RAR_EX
    rcall   debug_showRegister
    1: rjmp 1b


/////////////////////////////////
/// TLB MISS HANDLERS
/////////////////////////////////
.pagemask: .int 0b11111111111111111111000000000000
itlbMiss:
    pushm   r0-r3,r4-r7,r8-r9
    mov     r2,0b100000000
    rjmp    tlbMiss
dtlbMissWrite:
dtlbMissRead:
    pushm   r0-r3,r4-r7,r8-r9
    mov     r2,0
tlbMiss:
    /* get page requested */ 
    mfsr    r3,AVR32_TLBEAR

    lddpc   r0,.pagemask        // keep only page number
    and     r3,r0
    sbr     r3,9                // set the valid flag
    mov     r1,r3
    mfsr    r0,AVR32_TLBEHI
    bfins   r1,r0,0,8           // get the ASID

    /* At this point:
     * r2 = itlb or not
     * r3 = pagenumber +asid=0
     * r1 = pagenumber +asid
     */

    /* Search through page table for corresponding entry */
    mfsr    r9,AVR32_PTBR
    mov     r0,8193             // page table size + 1
searchPages:
    sub     r0,1
    breq    pageNotFound
    ld.w    r8,r9++
    ld.w    r5,r9++
    bld     r5,8
    breq    1f                  // page is global, don't check ASID    
    cp.w    r8,r1               // compare with a valid page with same ASID and VPN
    brne    searchPages         // entry found ?
    rjmp    pagefound
1:  cp.w    r8,r3               // global pages in page table will have ASID=0
    brne    searchPages
pagefound:
    or      r1,r2               // if it is a itlb miss, set the 'i' flag
    sbr     r5,1                // make sure dirty bit is set. We don't want an exception just for that

    /* load into tlb */
    mtsr    AVR32_TLBEHI,r1
    mtsr    AVR32_TLBELO,r5
    tlbw

    popm   r0-r3,r4-r7,r8-r9
    rete
pageNotFound:
    /* Page not found */
    SERIALPRINTF("\nTLB MISS ERROR: ")
    mfsr    r8,AVR32_TLBEAR
    rcall   debug_showRegister
    SERIALPRINTF(" From: ")
    mfsr    r8,AVR32_RAR_EX
    rcall   debug_showRegister
2:  rjmp   2b                   // nothing we can do.






////////////////////////////////
/// init_mmu
/// WARNING: Stack will be re-initialized and r9 will be changed
///////////////////////////////
.firstTLBEHI: .int 0x04000200   //first SRAM virtual page, valid dtlb, asid=0
.firstTLBELO: .int 0x24000327   // first SRAM physical page, cached,global,non-buffered,priviledged,4K,dirty, write-through
.firstSDRAMTLBELO: .int 0x10002277    // Third SDRAM physical page (1st two reserved for page table
init_mmu:
    stm     --sp,r7,r8,r9,r10,lr
    /* Build pages for sram */
    lddpc   r7,.firstTLBEHI
    lddpc   r8,.firstTLBELO
    lddpc   r9,pageTableBase

    mtsr    AVR32_PTBR,r9
    mtsr    AVR32_TLBEHI,r7     // load first sram page now (stack)
    mtsr    AVR32_TLBELO,r8     // load first sram page now (stack)

    mov     r10,8
1:  st.w    r9++,r7
    st.w    r9++,r8
    sub     r7,-0x1000          // increase page number
    sub     r8,-0x1000          // increase page number
    sub     r10,1
    brne    1b

    // at this point, r9 contains the next table entry, don't change it.
    /* Build pages for SDRAM */
    lddpc   r8,.firstSDRAMTLBELO
    mov     r7,0x0              // unallocated page. leave empty
    mov     r10,8176            // 8176 4k pages are availabl7
1:  st.w    r9++,r7
    st.w    r9++,r8
    sub     r8,-0x1000           // increase page number
    sub     r10,1
    brne    1b


    // WARNING: NOT SAFE TO USE STACK 
    ldm     sp++,r7,r8,r9,r10,lr
    lddpc   sp,stackBase        // reload new stack base address. All data will be lost

    /* Init MMU */
    /* We lock down first DTLB entry because it will be the stack page*/
    mov     r9,0b00010101      //paging, private, segmentation
    mtsr    AVR32_MMUCR,r9
    tlbw
    mov     r9,0b100000100010001                // DLA = 1
    mtsr    AVR32_MMUCR,r9
    // WARNING: SAFE TO USE STACK AGAIN

    mov     pc,lr               // return


pageTableBase: .int 0xB0000000
stackBase:     .int STACK_BASE

//////////////////////////////////////////////////////
/// Allocate one page of memory to a process
/// r8 = asid
/// return r12=virtual address
//////////////////////////////////////////////////////
PageAlloc:
    stm     --sp,r0,r6,r7,r9,r10,lr
    
    mov r0,0b1111111111
    /* Find next virtual address for this process */
    mov     r9,0
    mov     r10,8176
    mfsr    r7,AVR32_PTBR    
    sub     r7,-(8*8)               // skip sram pages
1:  ld.w    r6,r7[0]                // get TLBEHI
    sub     r7,-8                   // go to next record
    cp.w    r6,0                    // empty record? 
    breq    2f          
    cp.b    r8,r6                   // same ASID?
    brne    2f
    andn    r6,r0
    cp.w    r6,r9                   // address is higher than what we have now?
    brlo    2f
    mov     r9,r6  
    sub     r9,-0x1000              // we will use next page

2:  sub     r10,1
    brne    1b


    /* Find empty physical  block, At this point, r9 contains the virtual address */
    mov     r10,8176
    mfsr    r7,AVR32_PTBR
    sub     r7,-(8*8)               // skip sram pages
1:  ld.w    r6,r7[0]                // get TLBEHI
    sub     r7,-8                   // go to next record
    cp.w    r6,0                    // empty record?
    breq    2f
    sub     r10,1
    brne    1b
    mov     r12,0
    sub     r12,1
    rjmp    exitAlloc
2:  sub     r7,8

    mov     r12,r9
    sbr     r9,9
    bfins   r9,r8,0,8
    st.w    r7[0],r9 

exitAlloc:
    ldm     sp++,r0,r6,r7,r9,r10,pc

//////////////////////////////////////////////////////
/// Deallocate one page of memory for a process
/// r8 = asid
/// r9 = virtual address
//////////////////////////////////////////////////////
PageDealloc:
    stm     --sp,lr
//TODO: will also need to remove pages from TLB
    ldm     sp++,pc



//////////////////////////////////////////////////////
/// Deallocate all pages for a process
/// r8 = asid
//////////////////////////////////////////////////////
PageDeallocAll:
    stm     --sp,lr
//TODO: will also need to remove pages from TLB
    ldm     sp++,pc


/********************************
*** r7: number of bytes to allocate
*** RETURN: r12=virtual address of start buffer
********************************/
malloc:
    stm     --sp,r8,r9,r10,lr

    /* determine number of 4k pages to allocate */
    lsr     r7,12    
    mov     r9,r7
    sub     r9,-1

    /* determine ASID */
    mfsr    r8,AVR32_TLBEHI
    andl    r8,0xFF,COH
    /* allocate pages */    

    mov     r10,0
1:  rcall   PageAlloc
    cp      r10,0
    moveq   r10,r12

    sub     r9,1
    brne    1b

    mov     r12,r10
    ldm     sp++,r8,r9,r10,pc





//////////////////////////////////////////////////////
/// Allocate pages of memory to a process for code segment in process initialization
/// r10 = number of pages to allocate
/// r8 = asid
/// return r12=virtual address (-1 if not enough memory)
//////////////////////////////////////////////////////
PageAllocCode:
    stm     --sp,r0,r1,r6,r7,r9,r10,lr

    mov     r9,0                    // virtual address base

    /* Find empty physical  block, At this point, r9 contains the virtual address */
    mov     r1,8176
    mfsr    r7,AVR32_PTBR
    sub     r7,-(8*8)               // skip sram pages
1:  ld.w    r6,r7[0]                // get TLBEHI
    sub     r7,-8                   // go to next record
    cp.w    r6,0                    // empty record?
    breq    2f

    /* end of table, exit */
    sub     r1,1
    brne    1b
    mov     r12,0
    sub     r12,1
    rjmp    exitAllocCode

    /* empty page found */
2:  mov     r6,r7
    sub     r6,8
    mov     r12,r9
    sbr     r12,9
    bfins   r12,r8,0,8
    st.w    r6[0],r12
    sub     r9,-4096
    sub     r10,1
    brne    1b                      // allocated all requested page?
    mov     r12,0

exitAllocCode:
    ldm     sp++,r0,r1,r6,r7,r9,r10,pc




//////////////////////////////////////////////////////
/// Allocate pages of memory to a process for data segment in process initialization
/// r10 = number of pages to allocate
/// r8 = asid
/// return r12=virtual address (-1 if not enough memory)
//////////////////////////////////////////////////////
PageAllocData:
    stm     --sp,r0,r1,r6,r7,r9,r10,lr

    mov     r9,0x0000
    orh     r9,0x1000

    /* Find empty physical  block, At this point, r9 contains the virtual address */
    mov     r1,8176
    mfsr    r7,AVR32_PTBR
    sub     r7,-(8*8)               // skip sram pages
1:  ld.w    r6,r7[0]                // get TLBEHI
    sub     r7,-8                   // go to next record
    cp.w    r6,0                    // empty record?
    breq    2f

    /* end of table, exit */
    sub     r1,1
    brne    1b
    mov     r12,0
    sub     r12,1
    rjmp    exitAllocData

    /* empty page found */
2:  mov     r6,r7
    sub     r6,8
    mov     r12,r9
    sbr     r12,9
    bfins   r12,r8,0,8
    st.w    r6[0],r12
    sub     r9,-4096
    sub     r10,1
    brne    1b                      // allocated all requested page?

    mov     r12,0

exitAllocData:
    ldm     sp++,r0,r1,r6,r7,r9,r10,pc


