#include <avr32/io.h>
#include "macro.h"

.global scallBase

/****************************
*** The scall table
****************************/
.align 4
scallBase:
scall0: .int unhandledScall 

/***********************************************************
*** FUNC: serialSendBuf
*** DESC: send content of buffer on serial port. Not using PDC, will block until operation completed.
*** INPUT: r8=buffer address.
*** INPUT: r5=buffer size
*** INPUT: r7=uart number (0-3).
***********************************************************/
scall1: .int serialSendBuf

/***********************************************************
*** FUNC: serialReceive
*** DESC: NOT IMPLEMENTED
***********************************************************/
scall2: .int serialReceive

/***********************************************************
*** FUNC: serialInit
*** DESC: initialize usart as serial port
*** INPUT: r8=uart number (0-3).
*** INPUT: r7=baudrate (DOES NOT WORKING. HARDCODED TO 115200 8N1)
***********************************************************/
scall3: .int serialInit

/***********************************************************
*** FUNC: serialSendChar
*** DESC: send a single char on serial port
*** INPUT: r7=uart number (0-3).
*** INPUT: r8=byte to send
***********************************************************/
scall4: .int serialSendChar

scall5: .int unhandledScall
scall6: .int unhandledScall
scall7: .int unhandledScall
scall8: .int unhandledScall
scall9: .int unhandledScall
scall10: .int testScall

/***********************************************************
*** FUNC: dataflash_readpage
*** DESC: reads a 1056byte page from dataflash using SPI0, CS0
*** DESC: will use PDC to download data into the buffer and, depending
*** DESC: on 'r5', will wait for PDC to complete. Will wait for RDY before starting
*** INPUT: r8=buffer to receive page data. needs to be a physical address.
*** INPUT: r7=page Number (0-8191)
*** INPUT: r5=0 for blocking, non-zero if not blocking
***********************************************************/
scall11: .int dataflash_readpage

/******************************************************************
*** FUNC: dataflash_waitRecv 
*** DESC: Wait for PDC to be done receiving data from flash.
*** DESC: Usefull when using dataflash_readpage in non-blocking mode
******************************************************************/
scall12: .int dataflash_waitRecv

/******************************************************************
*** FUNC: dataflash_writepage
*** DESC: This will erase and write a full 1056byte page to dataflash.
*** DESC: will use PDC. Will not wait for status register to indicate RDY before returning
*** DESC: but will wait for RDY before starting.
*** INPUT: r8=buffer to send. must be in P1 or P2
*** INPUT: r7=page Number to write to
*** INPUT: r5=dont write if page has not changed (0=dont, 1=do it anyways)
*** RETURN: r12=number of bytes copied (0 if page was unchanged and r5==0)
******************************************************************/
scall13: .int dataflash_writepage
scall14: .int 0
scall15: .int 0

/******************************************************************
*** FUNC: malloc
*** DESC: allocate memory to process
*** INPUT: r7=number of bytes to allocate
*** RETURN: r12=buffer start address
******************************************************************/
scall16: .int malloc

scall17: .int 0
scall18: .int 0
scall19: .int 0


/******************************************************************
*** FUNC: fls
*** DESC: retrieve list of files in a directory
*** DESC: will return a list of 14bytes entries: 12char for 8.3 filename, 
*** DESC: 1 char for file type (1=file, 2=dir), and a null char
*** INPUT: r6=maximum number of entries to return
*** INPUT: r7=path
*** INPUT: r8=pointer to buffer to hold all entries
*** INPUT: r9=number of entries to skip
*** RETURN: r12=number of entries read
******************************************************************/
scall20: .int fls

/******************************************************************
*** FUNC: fopen
*** DESC: open file
*** INPUT: r8=pointer to an empty FILE structure to be filled by this function
*** INPUT: r7=path
*** RETURN: r12=0 if success
*****************************************************************/
scall21: .int fopen

/******************************************************************
*** FUNC: fread
*** DESC: read file
*** INPUT: r8=buffer in which to copy read data
*** INPUT: r11=number of bytes to read
*** INPUT: r5=pointer to a FILE structure initialized by fopen
*** RETURN: r12=number of bytes read
******************************************************************/
scall22: .int fread

/******************************************************************
*** FUNC: fclose
*** DESC: close file
*** INPUT: r5=pointer to a FILE structure initialized by fopen
******************************************************************/
scall23: .int fclose

/******************************************************************
*** FUNC: macb_getframe
*** DESC: receive an ethernet frame
*** INPUT: r5=MACB number
*** INPUT: r9=buffer to hold received frame (must be at least 1536 bytes)
*** RETURN: r12=number of bytes received
******************************************************************/
scall24: .int macb_getframe



unhandledScall:
    SERIALPRINTF("Unhandled scall\r\n")
    1:rjmp 1b

testScall:
    stm --sp,r8,lr
    
    mfsr    r8,AVR32_TLBEHI
    CLI
    SERIALPRINTF("\r\nscall: ")
    rcall   debug_showRegister
    STI
    ldm sp++,r8,pc
