#include <avr32/io.h>
#include "macro.h"

.global serialSendBuf
.global serialReceive
.global serialInit
.global serialSendChar

//////////////////////////////////////////////
/// serialInit
/// INPUT: r8= uart
/// INPUT: r7= baudrate
/// WARNINIG: Hardcoded to 8N1 115200
/////////////////////////////////////////////

serialInit:
    stm     --sp,r0,r9,lr
    lddpc   r6,.usart0
    mov     r0,r8
    lsl     r8,10
    add     r8,r8,r6

    /* reset usart */
    mov     r9,0
    st.w    r8[AVR32_USART_RTOR],r9
    st.w    r8[AVR32_USART_TTGR],r9
    com     r9
    st.w    r8[AVR32_USART_IDR],r9

    /* set baud rate */
    lddpc   r6,.baudRate0
    add     r6,r6,r7<<2
    //TODO: fix this. should not be hardcoded
    mov     r6,20                               //PBA=PLL0/4. baud=PBA/(16*20)
    st.w    r8[AVR32_USART_BRGR],r6


    /* set 8N1 */
    mov     r6,(((8-5)<< AVR32_USART_MR_CHRL_OFFSET) | \
                (0 << AVR32_USART_MR_CHMODE_OFFSET) | \
                (4 <<AVR32_USART_MR_PAR_OFFSET) | \
                (0 << AVR32_USART_MR_NBSTOP_OFFSET))
    st.w    r8[AVR32_USART_MR],r6

    /* enable interrupts */
    mov     r6,(1<<AVR32_USART_RXRDY)
    st.w    r8[AVR32_USART_IER],r6

    /* enable */
    mov     r6,(1<<AVR32_USART_CR_TXEN_OFFSET) | (1<<AVR32_USART_CR_RXEN_OFFSET)
    st.w    r8[AVR32_USART_CR],r6

    /* setup pio */
    cp      r0,0
    brne    chkU1
    lddpc   r7,.pioa
    mov     r9,((1<<8)|(1<<9))
    st.w    r7[AVR32_PIO_BSR],r9
    rjmp    setPIO
chkU1:
    cp      r0,1
    brne    chkU2
    lddpc   r7,.pioa
    mov     r9,((1<<AVR32_USART1_RXD_0_PIN)|(1<<AVR32_USART1_TXD_0_PIN))
    st.w    r7[AVR32_PIO_ASR],r9
    rjmp    setPIO
chkU2:
    cp      r0,2
    brne    chkU4
    lddpc   r7,.piob
    mov     r9,0
    sbr     r9,26
    sbr     r9,27
    st.w    r7[AVR32_PIO_BSR],r9
    rjmp    setPIO
chkU4:
    lddpc   r7,.piob
    mov     r9,((1<<17)|(1<<18))
    st.w    r7[AVR32_PIO_BSR],r9
setPIO:
    st.w    r7[AVR32_PIO_PDR],r9
    st.w    r7[AVR32_PIO_PUDR],r9

    ldm     sp++,r0,r9,pc

.pioa:          .int AVR32_PIOA_ADDRESS
.piob:          .int AVR32_PIOB_ADDRESS
.usart0:        .int AVR32_USART0_ADDRESS
.usartclk:      .int USART_CLK
//TODO: this table wont work
.baudRate0:     .int (2400/(16*USART_CLK))
.baudRate1:     .int (4800/(16*USART_CLK))
.baudRate2:     .int (9600/(16*USART_CLK))
.baudRate3:     .int (14400/(16*USART_CLK))
.baudRate4:     .int (19200/(16*USART_CLK))
.baudRate5:     .int (38400/(16*USART_CLK))
.baudRate6:     .int (57600/(16*USART_CLK))
.baudRate7:     .int 41


/**********************
*** serialSendBuf
*** INPUT: r8 = buf addr
*** INPUT: r5 = buf len
*** INPUT: r7 = usart number
**********************/
serialSendBuf:
    stm     --sp,r4,r6,r8,r9,lr

    cp.w    r5,0
    breq    3f

    lddpc   r9,.usart0
    lsl     r7,10
    add     r9,r9,r7


1:  ld.w    r6,r9[AVR32_USART_CSR]
    bld     r6,AVR32_USART_CSR_TXRDY_OFFSET
    brne    1b
    ld.ub   r4,r8++
    st.w    r9[AVR32_USART_THR],r4
    sub     r5,1
    brne    1b

3:  ldm     sp++,r4,r6,r8,r9,pc



/**********************
*** sendSerial
*** INPUT: r8 = byte to send
*** INPUT: r7 = usart  
**********************/
serialSendChar:
    stm     --sp,r6,r9,lr

    lddpc   r9,.usart0
    lsl     r7,10
    add     r9,r9,r7
    
trySend:
    ld.w    r6,r9[AVR32_USART_CSR]
    bld     r6,AVR32_USART_CSR_TXRDY_OFFSET
    brne    trySend     // if the bit is not set, try again
    
    st.w    r9[AVR32_USART_THR],r8

    ldm     sp++,r6,r9,pc



/***************************************************
*** recvSerial - Retrieve bytes from RX queue
*** RETURN: r12 = number of bytes read 
*** INPUT:  r8 = buffer to copy bytes to
*** INPUT:  r5 = max number of bytes to retrieve
*** INPUT:  r7 = usart number
*****************************************************/
serialReceive:
    stm     --sp,r8,lr

    ldm     sp++,r8,pc


